% Kernel Rank SVM training function using LASVM algorithm.
%
% Function parameters:
% ====================
% train_pts: A matrix of training points. Each column of the matrix is one point
% and each row is a feature dimension.
%
% orderings: A sparse matrix of point pairwise orderings with each row specifying
% ordering for one such pair. The number of columns must equal the total number
% of points in the training set. On each row, the column corresponding to the
% index (within train_pts) of the higher ranked point in the order is set to 1
% and the column corresponding to the index (within train_pts) of the lower
% ranked point is set to -1. All other columns must be set to 0.
%
% train_labels: A vector specifying whether an ordering is true or reversed
% within the ground truth orderings. The length of the vector is equal to the
% number of rows in the orderings matrix. A value of 1 indicates that the
% ordering implied by the corresponding row of the orderings matrix follows the
% known orderign in the ground truth data. A value of -1 indicates that the
% corresponding order in the orderings matrix is reversed w.r.t. the ground
% truth data. This vector is usually constructed by the calling function by
% randomly reversing the orders of certain pairs w.r.t. the ground truth when
% filling in the orderings matrix and then setting the corresponding entry in
% the train_labels to -1. This preserves the ground truth ordering while
% ensuring both positive and negative labels for the classifier.
%
% lambda: SVM Regularization parameter.
%
% kernel: A string specifying the name of the kernel function to be used. Other
% parameters specific to the kernel are passed through the varargin function
% argument. 
%
% n_epochs: Number of epochs through the training data.
%
% tau: Tolerance parameter of the LASVM algorithm.
%
% max_num_sv: Maximum number of support vectors to be retained. This takes
% effect if budget learning is enabled (do_budget). If kernel caching is used
% (see do_kernel_cache below), kernel cache is allocated to hold the kernel
% values of only max_num_sv orderings.
%
% do_squared_error: Sum of the square of the slack variable values is used in
% the objective function error term when this parameter is set to 1. If set to
% 0, the error term of the objective function is just the sum of the slack
% variable values.
%
% do_budget: Budget learning is enabled, i.e., the total number of support
% vectors learned is less than or equal to max_num_sv. One of a set of support
% vector pruning methods is used to eliminate excessive support vectors.
%
% elim_method: A string specifying the support vector pruning method to use to
% eliminate support vectors in excess of max_num_sv. This can be one of
% 'pairwise_gradient', 'single_gradient', or
% 'single_gradient_with_redistribution'.
%
% z_normalize: Data is normalized to have zero sample mean and unit sample
% variance when this argument is set to 1. Otherwise the data is used as is.
%
% use_sigmoid: Generate probabilistic output when plotting data. This really
% takes effect when plot_boundary is enabled. Otherwise, this option has no
% effect on training.
%
% compute_objective: Compute objective function value at each iteration. When
% enabled it significantly increases computation. Enable (set to 1) it only for
% debugging purposes.
%
% plot_boundary: It generates a plot of the learned function on a 2D grid of
% points. Mkaes sense only when the feature dimensionality is 2. It was
% implemented originally for debugging. It is usually disabled (set to 0).
%
% model_mat_file: Name of a Matlab .mat file to store the learned model
% parameters.
%
% do_kernel_cache: If set to 1, a cache for the rank-pairwise kernel matrix is
% used instead of computing the entire kernel matrix. This serves as a trade-off
% between memory and computational time.
%
% do_kernel_cache_check: If set to 1, the existence of kernel function value for
% a pair of orderings is prechecked before a call to the
% kernel_mat_pairwise_cache_fn(). This saves time when the size of the allocated
% cache is large but is observed to lead to overheads when the cache size is
% small. If set to 0, kernel_mat_pairwise_cache_fn() is always called for every
% pair.
% 
% do_point_pairwise_kernel_mat: If set to 1, kernel matrix between pairs of
% points is computed instead of between ordering pairs. This makes less use of
% memory at the cost of increasing computation in every epoch. If set to 0, the
% ordering pairwise kernel matrix is computed.
%
% do_sparse_kernel_mat: If set to 1, the kernel matrix is a sparse matrix, i.e.
% only the lower diagonal is computed and stored. If set to 0, the full matrix
% is computed and stored, which is perhaps inefficient in general.
%
% d_sv_preallocation: If set to 1, vectors are pre-allocated to store the
% Lagrange multiplier values, gradients and the indices of support vectors.
% Otherwise the vectors are grown or shrunk in size as support vectors are added
% to or deleted from the existing set, which is time-intensive. On the other
% hand, preallocation is probably slightly more memory intensive. 
%
% varargin: Argument used to specify kernel function specific parameters. For
% example, for the Gaussian kernel ('gauss_kernel'), scale parameter is
% specified using this argument. In the case of a polynomial kernel
% ('poly_kernel'), the polynomial degree and bias are specified using this
% argument.
%
% Function outputs:
% =================
% model: A structure 'model' with the following member variables. Optionally, it
% is possible to write the structure into a model mat file, which is specified
% as an input parameter. If the filename is empty, then the writing to disk is
% disabled.
% model.mean_pts: Mean for z normalization (vector of zeros, otherwise)
% model.std_pts: Standard deviation for z normalization (vector of ones, otherwise)
% model.sv_idxs: Indices of support vectors  
% model.sv_pts: Points corresponding to support vectors
% model.sv_alphas: Support vector coefficients
% model.sv_labels: Support vector class labels
% model.bias: Learned bias
% model.kernel: Kernel function string
% model.gauss_sigma: Gaussian kernel scale if Gaussian kernel is used
% model.poly_degree: Polynomial kernel degree if polynomial kernel is used
% model.poly_bias: Polynomial kernel bias if polynomial kernel is used
%
% NOTES:
% ======
% This function declares kernel_mat as a global variable, which might be
% accessible to other higher level calling functions such as
% ranksvm_kernel_sgd(). Once the kernel matrix is computed in that function, it
% is not recomputed again here. This is helpful because kernel matrix
% computation is expensive, and can be reused for multiple runs, eg. parameter
% search.

function [model] = kernel_rank_lasvm_sgd_train(train_pts, ...
                                               orderings, ...
                                               train_labels, ...
                                               lambda, ...
                                               kernel, ...
                                               n_epochs, ...
                                               tau, ...
                                               max_num_sv, ...
                                               do_squared_error, ...
                                               do_budget, ...
                                               elim_method, ...
                                               z_normalize, ...
                                               use_sigmoid, ...
                                               compute_objective, ...
                                               plot_boundary, ...
                                               model_mat_file, ...
                                               do_kernel_cache, ...
                                               do_kernel_cache_check, ...
                                               do_point_pairwise_kernel_mat, ...
                                               do_sparse_kernel_mat, ...
                                               do_sv_preallocation, ...
                                               varargin)

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;
global bias;
global delta;

global pts;
global rank_orderings;
global orderings_pt_idxs;
global labels;
global box_constraint;
global kernel_mat;
global kernel_cache;

global gauss_sigma;
global poly_degree;
global poly_bias;

global doing_kernel_cache;
global doing_kernel_cache_check;
global doing_pt_pairwise_kernel_mat;
global doing_sparse_kernel_mat;
global doing_sq_error;
global doing_sv_preallocation;

global kernel_value;

if (~isempty(train_pts))
    pts    = train_pts;
    labels = train_labels;
else
    %pts    = [[1; 1] [1; -1] [-1; -1] [-1; 1]]
    %pts    = [[0; 1] [1; 0] [0; -1] [-1; 0]]
    %pts    = pts + [[0; 1] [0; 1] [0; 1] [0; 1]]
    %labels = [1; -1; 1; -1];

    % Random data.
    rng(466456);
    n_pts = 5;

    pts   = zeros(2, n_pts);
    for i = 1:n_pts;
        pts(1, i) = 2*rand - 1;
        pts(2, i) = 2*rand - 1;
    end

    n_O = size(orderings, 1);
    labels = ones(n_O, 1);
    for i = 1:n_O
        if (rand < 0.5)
            labels(i) = -1;
        end
    end
end

% Copy parameters into globals.
rank_orderings               = orderings;
doing_kernel_cache           = do_kernel_cache;
doing_kernel_cache_check     = do_kernel_cache_check;
doing_pt_pairwise_kernel_mat = do_point_pairwise_kernel_mat;
doing_sparse_kernel_mat      = do_sparse_kernel_mat;
doing_sq_error               = do_squared_error;
doing_sv_preallocation       = do_sv_preallocation

n_pts = size(pts, 2);
n_dim = size(pts, 1);
n_O   = size(orderings, 1);
box_constraint = 1 / lambda;

% Precompute point idxs of the orderings matrix.
orderings_pt_idxs = zeros(n_O, 2);
for o_i = 1:n_O
    orderings_pt_idxs(o_i, 1) = find(orderings(o_i, :) == 1);
    orderings_pt_idxs(o_i, 2) = find(orderings(o_i, :) == -1);
end

% Z normalization.
mean_pts = zeros(n_dim, 1);
std_pts  = ones(n_dim, 1);
if(z_normalize == 1)
    disp('Performing unit normalization.............');
    tic
    mean_pts = mean(pts, 2);
    std_pts  = std(pts, 1, 2);

    % Handle zero standard deviations.
    singular_dims = find(std_pts < eps);
    disp('WARNING: Singular dimensions exist, skipping Z transformation for such dimensions!');
    std_pts(singular_dims) = 1;

    % In theory, the following is slow.
    %pts      = pts - repmat(mean_pts, [1 n_pts]);
    %pts      = pts ./ repmat(std_pts, [1 n_pts]);

    pts = pts - mean_pts(:, ones(n_pts, 1));
    pts = pts ./ std_pts(:, ones(n_pts, 1));
    toc
end

% Debugging.
%mean(pts, 2);
%std(pts, 1, 2);

if (nargin > 0)
    if (~isempty(strfind(kernel, 'gauss')))
        % Gaussian kernel:
        gauss_sigma = varargin{1};
    elseif (~isempty(strfind(kernel, 'poly')))
        % Polynomial kernel:
        poly_degree = varargin{1};
        poly_bias   = varargin{2};
    end
else
    % Gaussian kernel:
    gauss_sigma = 1.0;
    
    % Polynomial kernel:
    poly_degree = 2;
    poly_bias   = 1;
end

% Build the kernel matrix.
if( isempty(kernel_mat) )
    disp('Computing kernel matrix.............');
    tic

    if (doing_pt_pairwise_kernel_mat)
        disp('Computing point-pairwise kernel matrix.............');

        if (doing_sparse_kernel_mat)
            disp('Computing sparse kernel matrix.............');
            kernel_mat = sparse([], [], n_pts, n_pts, (n_pts*(n_pts+1)/2));
            for i = 1:n_pts
                for j = 1:i        
                    if (~isempty(strfind(kernel, 'gauss')))
                        kernel_mat(i, j) = feval(kernel, pts(:, i), pts(:, j), gauss_kernel_scale);
                    elseif (~isempty(strfind(kernel, 'poly')))
                        kernel_mat(i, j) = feval(kernel, pts(:, i), pts(:, j), poly_degree, poly_bias);
                    else
                        kernel_mat(i, j) = feval(kernel, pts(:, i), pts(:, j));
                    end
                end
            end
        else
            disp('Computing full kernel matrix.............');
            kernel_mat = zeros(n_pts, n_pts);
            for i = 1:n_pts
                for j = 1:n_pts        
                    if (~isempty(strfind(kernel, 'gauss')))
                        kernel_mat(i, j) = feval(kernel, pts(:, i), pts(:, j), gauss_kernel_scale);
                    elseif (~isempty(strfind(kernel, 'poly')))
                        kernel_mat(i, j) = feval(kernel, pts(:, i), pts(:, j), poly_degree, poly_bias);
                    else
                        kernel_mat(i, j) = feval(kernel, pts(:, i), pts(:, j));
                    end
                end
            end
        end

    else
        disp('Computing rank-pairwise kernel matrix.............');

        if (doing_sparse_kernel_mat)
            disp('Computing sparse kernel matrix.............');
            kernel_mat = sparse([], [], [], n_O, n_O, (n_O*(n_O+1)/2));
            for o_1 = 1:n_O
                for o_2 = 1:o_1
                    i = find(orderings(o_1,:) == 1)
                    j = find(orderings(o_1,:) == -1)
                    k = find(orderings(o_2,:) == 1)
                    l = find(orderings(o_2,:) == -1)

                    if (~isempty(strfind(kernel, 'gauss')))
	                kernel_mat(o_1, o_2) = (1 + feval(kernel, pts(:, i), pts(:, k), gauss_sigma) ...
                                                  - feval(kernel, pts(:, i), pts(:, l), gauss_sigma) ...
                                                  - feval(kernel, pts(:, j), pts(:, k), gauss_sigma) ...
                                                  + feval(kernel, pts(:, j), pts(:, l), gauss_sigma)    );                

                    elseif (~isempty(strfind(kernel, 'poly')))
	                kernel_mat(o_1, o_2) = (1 + feval(kernel, pts(:, i), pts(:, k), poly_degree, poly_bias) ...
                                                  - feval(kernel, pts(:, i), pts(:, l), poly_degree, poly_bias) ...
                                                  - feval(kernel, pts(:, j), pts(:, k), poly_degree, poly_bias) ...
                                                  + feval(kernel, pts(:, j), pts(:, l), poly_degree, poly_bias)    );
                    else
	                kernel_mat(o_1, o_2) = (1 + feval(kernel, pts(:, i), pts(:, k)) ...
                                                  - feval(kernel, pts(:, i), pts(:, l)) ...
                                                  - feval(kernel, pts(:, j), pts(:, k)) ...
                                                  + feval(kernel, pts(:, j), pts(:, l))    );
                    end

                    if (doing_sq_error)
                        if (o_1 == o_2)
                            kernel_mat(o_1, o_2) = kernel_mat(o_1, o_2) + lambda;
                        end
                    end
                end
            end
        else
            disp('Computing full kernel matrix.............');
            kernel_mat = zeros(n_O, n_O);
            for o_1 = 1:n_O
                for o_2 = 1:n_O
                    i = find(orderings(o_1,:) == 1)
                    j = find(orderings(o_1,:) == -1)
                    k = find(orderings(o_2,:) == 1)
                    l = find(orderings(o_2,:) == -1)

                    if (~isempty(strfind(kernel, 'gauss')))
	                kernel_mat(o_1, o_2) = (1 + feval(kernel, pts(:, i), pts(:, k), gauss_sigma) ...
                                                  - feval(kernel, pts(:, i), pts(:, l), gauss_sigma) ...
                                                  - feval(kernel, pts(:, j), pts(:, k), gauss_sigma) ...
                                                  + feval(kernel, pts(:, j), pts(:, l), gauss_sigma)    );                

                    elseif (~isempty(strfind(kernel, 'poly')))
	                kernel_mat(o_1, o_2) = (1 + feval(kernel, pts(:, i), pts(:, k), poly_degree, poly_bias) ...
                                                  - feval(kernel, pts(:, i), pts(:, l), poly_degree, poly_bias) ...
                                                  - feval(kernel, pts(:, j), pts(:, k), poly_degree, poly_bias) ...
                                                  + feval(kernel, pts(:, j), pts(:, l), poly_degree, poly_bias)    );
                    else
	                kernel_mat(o_1, o_2) = (1 + feval(kernel, pts(:, i), pts(:, k)) ...
                                                  - feval(kernel, pts(:, i), pts(:, l)) ...
                                                  - feval(kernel, pts(:, j), pts(:, k)) ...
                                                  + feval(kernel, pts(:, j), pts(:, l))    );
                    end

                    if (doing_sq_error)
                        if (o_1 == o_2)
                            kernel_mat(o_1, o_2) = kernel_mat(o_1, o_2) + lambda;
                        end
                    end
                end
            end
        end
        
    end
    toc
else
    disp('Using precomputed kernel matrix.............');
end
% Debugging.
%pts
%orderings
%labels
%kernel_mat

if (doing_kernel_cache)
    disp('Initializing kernel cache.............');
    kernel_cache.max_size    = max_num_sv;
    kernel_cache.curr_size   = 0;
    kernel_cache.idx_to_row  = zeros(n_O, 1);
    kernel_cache.row_to_idx  = zeros(max_num_sv, 1);
    kernel_cache.matrix      = zeros(max_num_sv, max_num_sv); 
end

% Initialization
if (doing_sv_preallocation)
    n_sv         = 0;
    sv_idxs      = zeros(n_O, 1);
    sv_labels    = zeros(n_O, 1);
    sv_alphas    = zeros(n_O, 1);
    sv_gradients = zeros(n_O, 1);
else
    sv_idxs      = [];
    sv_labels    = [];
    sv_alphas    = [];
    sv_gradients = [];
end

% Pre-allocate variables to hold kernel values.
kernel_value = 0;
kv_ij        = 0;

objective_fn  = zeros(n_epochs*n_O, 1);
constraint_fn = zeros(n_epochs*n_O, 1);
t = 1;
%done = 0;
for epoch = 1:n_epochs

    % Debugging.
    disp(strcat('Epoch: ', num2str(epoch)));
    tic
    rand_idxs = randperm(n_O);

    for i = 1:n_O
        %disp('Process.............')
        lasvm_process(rand_idxs(i), tau);

        % Debugging.
        %sv_idxs
        %sv_labels
        %sv_alphas
        %sv_gradients

        %disp('Reprocess.............')
        lasvm_reprocess(tau);

        % Debugging.
        %sv_idxs
        %sv_labels
        %sv_alphas
        %sv_gradients

        if (do_budget == 1)
            reoptimize = 1;

            if ( (~isempty(strfind(elim_method, 'single'))) & (~isempty(strfind(elim_method, 'redistribution'))))
                eliminate_svs_single_gradient_with_redistribution(max_num_sv, tau, reoptimize);
            elseif ( ~isempty(strfind(elim_method, 'single')) )
                eliminate_svs_single_gradient(max_num_sv, tau, reoptimize);
            else
                eliminate_svs_pairwise_gradient(max_num_sv, tau, reoptimize);
            end        
        end

        % Debugging.
        %sv_idxs
        %sv_labels
        %sv_alphas
        %sv_gradients
       
        %if (early_exit == 1)
        %    n_sv = length(sv_idxs);
        %
        %    if (n_sv >= max_num_sv)
        %        done = 1;
        %        break;
        %    end
        %end
       
        if (compute_objective == 1)
            if (doing_sv_preallocation)
                sv_labeled_alphas = sv_labels(1:n_sv) .* sv_alphas(1:n_sv);
            else
                sv_labeled_alphas = sv_labels .* sv_alphas;
            end
            if (doing_pt_pairwise_kernel_mat)
                if (~doing_sv_preallocation)
                    n_sv = length(sv_idxs);
                end
                for sv_i = 1:n_sv
                    for sv_j = 1:n_sv
                        if (doing_kernel_cache)
                            %objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas(sv_i) * kernel_mat_pairwise_cache_fn(sv_idxs(sv_i), sv_idxs(sv_j)) * sv_labeled_alphas(sv_j)));

                            %kv_ij           = kernel_mat_pairwise_cache_fn(sv_idxs(sv_i), sv_idxs(sv_j));

                            %if (doing_kernel_cache_check)
                            %    idx_to_row_i = kernel_cache.idx_to_row(sv_idxs(sv_i));
                            %    idx_to_row_j = kernel_cache.idx_to_row(sv_idxs(sv_j));
                            %    if ( (idx_to_row_i > 0) && (idx_to_row_j > 0) )
                            %        if (idx_to_row_i > idx_to_row_j)
                            %            kv_ij = kernel_cache.matrix(idx_to_row_i, idx_to_row_j);
                            %        else
                            %            kv_ij = kernel_cache.matrix(idx_to_row_j, idx_to_row_i);
                            %        end
                            %    else
                            %        kernel_mat_pairwise_cache_fn(sv_idxs(sv_i), sv_idxs(sv_j));
                            %        kv_ij = kernel_value;
                            %    end
                            %else
                                kernel_mat_pairwise_cache_fn(sv_idxs(sv_i), sv_idxs(sv_j));
                                kv_ij = kernel_value;
                            %end

                            objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas(sv_i) * kv_ij * sv_labeled_alphas(sv_j)));
                        else
                            objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas(sv_i) * kernel_mat_pairwise_fn(sv_idxs(sv_i), sv_idxs(sv_j)) * sv_labeled_alphas(sv_j)));
                        end
                    end
                end
            else
                if (doing_sparse_kernel_mat)
                    if (doing_sv_preallocation)
                        objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas' * kernel_mat(sv_idxs(1:n_sv), sv_idxs(1:n_sv)) * sv_labeled_alphas)); 
                        objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas' * kernel_mat(sv_idxs(1:n_sv), sv_idxs(1:n_sv))' * sv_labeled_alphas));
                        objective_fn(t) = objective_fn(t) - sum((sv_alphas(1:n_sv).^2).*diag(kernel_mat(sv_idxs(1:n_sv), sv_idxs(1:n_sv))));  
                    else
                        objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas' * kernel_mat(sv_idxs, sv_idxs) * sv_labeled_alphas)); 
                        objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas' * kernel_mat(sv_idxs, sv_idxs)' * sv_labeled_alphas));
                        objective_fn(t) = objective_fn(t) - sum((sv_alphas.^2).*diag(kernel_mat(sv_idxs, sv_idxs)));  
                    end
                else
                    if (doing_sv_preallocation)
                        objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas' * kernel_mat(sv_idxs(1:n_sv), sv_idxs(1:n_sv)) * sv_labeled_alphas));
                    else
                        objective_fn(t) = objective_fn(t) + (0.5 * (sv_labeled_alphas' * kernel_mat(sv_idxs, sv_idxs) * sv_labeled_alphas));
                    end
                end
            end
            if (doing_sv_preallocation)
                objective_fn(t)   = objective_fn(t) - sum(sv_alphas(1:n_sv));
            else
                objective_fn(t)   = objective_fn(t) - sum(sv_alphas);
            end

            constraint_fn(t) = sum(sv_labeled_alphas);
        end
        t = t + 1; 
    end

    %if (done == 1)
    %    break;
    %end

    toc
end

% Finishing
lasvm_finish_reprocess(tau);

if (plot_boundary == 1)
    plot_decision_boundary(kernel, orderings, use_sigmoid);
end

%if (early_exit == 0)
% Eliminating least useful SVs

if (do_budget == 0)
    reoptimize = 1;

    if ( (~isempty(strfind(elim_method, 'single'))) & (~isempty(strfind(elim_method, 'redistribution'))))
        eliminate_svs_single_gradient_with_redistribution(max_num_sv, tau, reoptimize);
    elseif ( ~isempty(strfind(elim_method, 'single')) )
        eliminate_svs_single_gradient(max_num_sv, tau, reoptimize);
    elseif ( ~isempty(strfind(elim_method, 'pairwise')) ) 
        eliminate_svs_pairwise_gradient(max_num_sv, tau, reoptimize);
    else
        % Do nothing.
    end
end

% Reoptimize after eliminating alphas
lasvm_finish_reprocess(tau);
%end

if (plot_boundary == 1)
    plot_decision_boundary(kernel, orderings, use_sigmoid);
end

% Plot the objective function.
figure, plot(objective_fn(:), 'k-');
xlabel('Iteration'), ylabel('Error'); 

% Plot the constraint function.
figure, plot(constraint_fn(:), 'k-');
xlabel('Iteration'), ylabel('Constraint');    

% Plot the alphas and their gradients.
figure, hold on;
plot((sv_labels .* sv_alphas), 'k-'),
plot(sv_gradients, 'r-'),
xlabel('SV Index'), ylabel('Alphas (Black) and Gradients(Red)');

% Return parameter(s).
if (~doing_sv_preallocation)
    n_sv = length(sv_alphas);
end

model.mean_pts = mean_pts;
model.std_pts  = std_pts;

if (doing_sv_preallocation)
    model.sv_idxs = sv_idxs(1:n_sv);
else
    model.sv_idxs = sv_idxs;
end

model.sv_pts = [];
for s = 1:n_sv
    high_idx = find(orderings(sv_idxs(s), :) == 1);
    low_idx  = find(orderings(sv_idxs(s), :) == -1);

    model.sv_pts = [model.sv_pts [pts(:, high_idx); pts(:, low_idx)]];
end

if (doing_sv_preallocation)
    model.sv_alphas   = sv_alphas(1:n_sv);
    model.sv_labels   = sv_labels(1:n_sv);
else
    model.sv_alphas   = sv_alphas;
    model.sv_labels   = sv_labels;
end
model.bias        = bias;
model.kernel      = kernel;
model.gauss_sigma = gauss_sigma;
model.poly_degree = poly_degree;
model.poly_bias   = poly_bias;

if(~isempty(model_mat_file))
    save(model_mat_file, 'model', '-mat');
end

clear pts;

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [] = lasvm_process(k, tau)

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;
global bias;
global delta;

global labels;
global kernel_mat;
global kernel_cache;
global box_constraint;

global doing_kernel_cache;
global doing_kernel_cache_check;
global doing_pt_pairwise_kernel_mat;
global doing_sparse_kernel_mat;
global doing_sq_error;
global doing_sv_preallocation;

global orderings_pt_idxs;

% Pre-allocate variables to hold kernel values.
global kernel_value;

kernel_value = 0;
kv_ii        = 0;
kv_jj        = 0;
kv_ij        = 0;
kv_is        = 0;
kv_js        = 0;
kv_ks        = 0;

if (~doing_sv_preallocation)
    n_sv = length(sv_idxs);
end

if (isempty(find(sv_idxs == k)))   
    % Add this point to the kernel cache, 
    % since gradient needs to computed anyway.
    if (doing_kernel_cache)
        if (doing_pt_pairwise_kernel_mat)
            kernel_cache_add_row_pairwise(k);
        end
    end

    idx_k   = k;
    y_k     = labels(idx_k);
    alpha_k = 0;
    wphi    = 0;
    for s = 1:n_sv    
        idx_s = sv_idxs(s);
        y_s   = sv_labels(s);
        if (doing_pt_pairwise_kernel_mat)    
            if (doing_kernel_cache)
                %wphi = wphi + (sv_alphas(s) * y_s * kernel_mat_pairwise_cache_fn(idx_k, idx_s));

                %kv_ks = kernel_mat_pairwise_cache_fn(idx_k, idx_s);

                %if (doing_kernel_cache_check)
                %    idx_to_row_k = kernel_cache.idx_to_row(idx_k);
                %    idx_to_row_s = kernel_cache.idx_to_row(idx_s);
                %    if ( (idx_to_row_k > 0) && (idx_to_row_s > 0) )
                %        if (idx_to_row_k > idx_to_row_s)
                %            kv_ks = kernel_cache.matrix(idx_to_row_k, idx_to_row_s);
                %        else
                %            kv_ks = kernel_cache.matrix(idx_to_row_s, idx_to_row_k);
                %        end
                %    else
                %        kernel_mat_pairwise_cache_fn(idx_k, idx_s);
                %        kv_ks = kernel_value;
                %    end
                %else
                    kernel_mat_pairwise_cache_fn(idx_k, idx_s);
                    kv_ks = kernel_value;
                %end
                wphi  = wphi + (sv_alphas(s) * y_s * kv_ks); 
            else        
                wphi = wphi + (sv_alphas(s) * y_s * kernel_mat_pairwise_fn(idx_k, idx_s));
            end
        else
            if (doing_sparse_kernel_mat)
                wphi = wphi + (sv_alphas(s) * y_s * kernel_mat(max([idx_k, idx_s]), min([idx_k, idx_s])));
            else
                wphi = wphi + (sv_alphas(s) * y_s * kernel_mat(idx_k, idx_s));
            end
        end
    end
    gradient_k = y_k - wphi;

    % Debugging.
    %n_sv

    % Add this point to the set of SVs
    sv_added = 0;
    %if (isempty(find(sv_labels == labels(k))))
        if (doing_sv_preallocation)
            sv_idxs(n_sv+1)      = k;
            sv_labels(n_sv+1)    = labels(k);
            sv_alphas(n_sv+1)    = alpha_k;
            sv_gradients(n_sv+1) = gradient_k;
        else
            sv_idxs      = [sv_idxs; k];
            sv_labels    = [sv_labels; labels(k)];
            sv_alphas    = [sv_alphas; alpha_k];
            sv_gradients = [sv_gradients; gradient_k];
        end

        sv_added = 1;
    %end
else
    return;
end

if (y_k > 0)
    sv_idx_i   = n_sv + 1;
    idx_i      = k;
    gradient_i = gradient_k;
    alpha_i    = alpha_k;

    [gradient_j, sv_idx_j, idx_j, alpha_j] = find_min_gradient_neg_class();
    if (isempty(idx_j))        
        if (doing_sv_preallocation)
            n_sv = n_sv + 1;
        end

        return;
    end
else
    sv_idx_j   = n_sv + 1;
    idx_j      = k;
    gradient_j = gradient_k;
    alpha_j    = alpha_k;

    [gradient_i, sv_idx_i, idx_i, alpha_i] = find_max_gradient_pos_class();
    if (isempty(idx_i))
        if (doing_sv_preallocation)
            n_sv = n_sv + 1;
        end

        return;
    end
end

gradient_diff = gradient_i - gradient_j; 
% Check if tau-violating
if (gradient_diff <= tau)
    if (doing_sv_preallocation)
        n_sv = n_sv + 1;
    end

    return;
else
    % Add this point to the set of SVs
    if (~sv_added)
        if (doing_sv_preallocation)
            sv_idxs(n_sv+1)      = k;
            sv_labels(n_sv+1)    = labels(k);
            sv_alphas(n_sv+1)    = alpha_k;
            sv_gradients(n_sv+1) = gradient_k;
        else
            sv_idxs      = [sv_idxs; k];
            sv_labels    = [sv_labels; labels(k)];
            sv_alphas    = [sv_alphas; alpha_k];
            sv_gradients = [sv_gradients; gradient_k];
        end
    end

    if (doing_pt_pairwise_kernel_mat)
        if (doing_kernel_cache)
            %hessian = kernel_mat_pairwise_cache_fn(idx_i, idx_i) + kernel_mat_pairwise_cache_fn(idx_j, idx_j) - (2*kernel_mat_pairwise_cache_fn(idx_i, idx_j));

            %kv_ii   = kernel_mat_pairwise_cache_fn(idx_i, idx_i)                       
            %kv_jj   = kernel_mat_pairwise_cache_fn(idx_j, idx_j)
            %kv_ij   = kernel_mat_pairwise_cache_fn(idx_i, idx_j)

            %if (doing_kernel_cache_check)
            %    idx_to_row_i = kernel_cache.idx_to_row(idx_i);
            %    if ( (idx_to_row_i > 0) )
            %        kv_ii = kernel_cache.matrix(idx_to_row_i, idx_to_row_i);             
            %    else
            %        kernel_mat_pairwise_cache_fn(idx_i, idx_i);
            %        kv_ii = kernel_value;                       
            %    end
            %    idx_to_row_j = kernel_cache.idx_to_row(idx_j);
            %    if ( (idx_to_row_j > 0) )
            %        kv_jj = kernel_cache.matrix(idx_to_row_j, idx_to_row_j);             
            %    else
            %        kernel_mat_pairwise_cache_fn(idx_j, idx_j);
            %        kv_jj = kernel_value;
            %    end
            %    if ( (idx_to_row_i > 0) && (idx_to_row_j > 0) )
            %        if (idx_to_row_i > idx_to_row_j)
            %            kv_ij = kernel_cache.matrix(idx_to_row_i, idx_to_row_j);
            %        else
            %            kv_ij = kernel_cache.matrix(idx_to_row_j, idx_to_row_i);
            %        end
            %    else
            %        kernel_mat_pairwise_cache_fn(idx_i, idx_j)
            %        kv_ij = kernel_value;
            %    end
            %else
                kernel_mat_pairwise_cache_fn(idx_i, idx_i);
                kv_ii = kernel_value;                       
                kernel_mat_pairwise_cache_fn(idx_j, idx_j);
                kv_jj = kernel_value;
                kernel_mat_pairwise_cache_fn(idx_i, idx_j)
                kv_ij = kernel_value;
            %end

            hessian = kv_ii + kv_jj - (2*kv_ij);
        else
            hessian = kernel_mat_pairwise_fn(idx_i, idx_i) + kernel_mat_pairwise_fn(idx_j, idx_j) - (2*kernel_mat_pairwise_fn(idx_i, idx_j));
        end
    else
        hessian = kernel_mat(idx_i, idx_i) + kernel_mat(idx_j, idx_j);
        if (doing_sparse_kernel_mat)
            hessian = hessian - (2*kernel_mat(max([idx_i, idx_j]), min([idx_i, idx_j])));
        else
            hessian = hessian - (2*kernel_mat(idx_i, idx_j));
        end
    end

    lambda = min([(gradient_diff/hessian), (box_constraint-alpha_i), (box_constraint-alpha_j)]);   
    sv_alphas(sv_idx_i) = sv_alphas(sv_idx_i) + lambda;
    sv_alphas(sv_idx_j) = sv_alphas(sv_idx_j) + lambda;

    if (doing_sv_preallocation)
        n_sv = n_sv + 1;
    else
        n_sv = length(sv_idxs);
    end
    for s = 1:n_sv
        if (doing_pt_pairwise_kernel_mat)
            if (doing_kernel_cache)
                %sv_gradients(s) = sv_gradients(s) - (lambda * (kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s)) - kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s))));

                %kv_is           = kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                %kv_js           = kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));

                %if (doing_kernel_cache_check)
                %    idx_to_row_i = kernel_cache.idx_to_row(idx_i);
                %    idx_to_row_s = kernel_cache.idx_to_row(sv_idxs(s));
                %    if ( (idx_to_row_i > 0) && (idx_to_row_s > 0) )
                %        if (idx_to_row_i > idx_to_row_s)
                %            kv_is = kernel_cache.matrix(idx_to_row_i, idx_to_row_s);
                %        else
                %            kv_is = kernel_cache.matrix(idx_to_row_s, idx_to_row_i);
                %        end
                %    else
                %        kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                %        kv_is = kernel_value;
                %    end
                %    idx_to_row_j = kernel_cache.idx_to_row(idx_j);
                %    if ( (idx_to_row_j > 0) && (idx_to_row_s > 0) )
                %        if (idx_to_row_j > idx_to_row_s)
                %            kv_js = kernel_cache.matrix(idx_to_row_j, idx_to_row_s);
                %        else
                %            kv_js = kernel_cache.matrix(idx_to_row_s, idx_to_row_j);
                %        end
                %    else             
                %        kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));
                %        kv_js = kernel_value;
                %    end
                %else
                    kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                    kv_is = kernel_value;
                    kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));
                    kv_js = kernel_value;
                %end

                sv_gradients(s) = sv_gradients(s) - (lambda * (kv_is - kv_js));
            else
                sv_gradients(s) = sv_gradients(s) - (lambda * (kernel_mat_pairwise_fn(idx_i, sv_idxs(s)) - kernel_mat_pairwise_fn(idx_j, sv_idxs(s))));
            end
        else
            if (doing_sparse_kernel_mat)
                sv_gradients(s) = sv_gradients(s) - (lambda * (kernel_mat(max([idx_i, sv_idxs(s)]), min([idx_i, sv_idxs(s)])) - kernel_mat(max([idx_j, sv_idxs(s)]), min([idx_j, sv_idxs(s)]))));
            else
                sv_gradients(s) = sv_gradients(s) - (lambda * (kernel_mat(idx_i, sv_idxs(s)) - kernel_mat(idx_j, sv_idxs(s))));
            end
        end
    end
end

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [] = lasvm_reprocess(tau)

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;
global bias;
global delta;

global kernel_mat;
global kernel_cache;
global box_constraint;

global doing_kernel_cache;
global doing_kernel_cache_check;
global doing_pt_pairwise_kernel_mat;
global doing_sparse_kernel_mat;
global doing_sq_error;
global doing_sv_preallocation;

global orderings_pt_idxs;

% Pre-allocate variables to hold kernel values.
global kernel_value;

kernel_value = 0;
kv_ii        = 0;
kv_jj        = 0;
kv_ij        = 0;
kv_is        = 0;
kv_js        = 0;
kv_ks        = 0;

[gradient_i, sv_idx_i, idx_i, alpha_i] = find_max_gradient_pos_class();    
if (isempty(idx_i))
    return;
end

[gradient_j, sv_idx_j, idx_j, alpha_j] = find_min_gradient_neg_class();    
if (isempty(idx_j))
    return;
end

gradient_diff = gradient_i - gradient_j;
delta         = gradient_diff; 
% Check if tau-violating
if (gradient_diff <= tau)
    return;
else
    if (doing_pt_pairwise_kernel_mat)
        if (doing_kernel_cache)
            %hessian = kernel_mat_pairwise_cache_fn(idx_i, idx_i) + kernel_mat_pairwise_cache_fn(idx_j, idx_j) - (2*kernel_mat_pairwise_cache_fn(idx_i, idx_j));

            %kv_ii   = kernel_mat_pairwise_cache_fn(idx_i, idx_i);
            %kv_jj   = kernel_mat_pairwise_cache_fn(idx_j, idx_j);
            %kv_ij   = kernel_mat_pairwise_cache_fn(idx_i, idx_j);

            %if (doing_kernel_cache_check)
            %    idx_to_row_i = kernel_cache.idx_to_row(idx_i);
            %    if ( (idx_to_row_i > 0) )
            %        kv_ii = kernel_cache.matrix(idx_to_row_i, idx_to_row_i);             
            %    else
            %        kernel_mat_pairwise_cache_fn(idx_i, idx_i);
            %        kv_ii = kernel_value;                       
            %    end
            %    idx_to_row_j = kernel_cache.idx_to_row(idx_j);
            %    if ( (idx_to_row_j > 0) )
            %        kv_jj = kernel_cache.matrix(idx_to_row_j, idx_to_row_j);             
            %    else
            %        kernel_mat_pairwise_cache_fn(idx_j, idx_j);
            %        kv_jj = kernel_value;
            %    end
            %    if ( (idx_to_row_i > 0) && (idx_to_row_j > 0) )
            %        if (idx_to_row_i > idx_to_row_j)
            %            kv_ij = kernel_cache.matrix(idx_to_row_i, idx_to_row_j);
            %        else
            %            kv_ij = kernel_cache.matrix(idx_to_row_j, idx_to_row_i);
            %        end
            %    else
            %        kernel_mat_pairwise_cache_fn(idx_i, idx_j)
            %        kv_ij = kernel_value;
            %    end
            %else
                kernel_mat_pairwise_cache_fn(idx_i, idx_i);
                kv_ii = kernel_value;                       
                kernel_mat_pairwise_cache_fn(idx_j, idx_j);
                kv_jj = kernel_value;
                kernel_mat_pairwise_cache_fn(idx_i, idx_j)
                kv_ij = kernel_value;
            %end

            hessian = kv_ii + kv_jj - (2*kv_ij);
        else
            hessian = kernel_mat_pairwise_fn(idx_i, idx_i) + kernel_mat_pairwise_fn(idx_j, idx_j) - (2*kernel_mat_pairwise_fn(idx_i, idx_j));
        end
    else
        hessian = kernel_mat(idx_i, idx_i) + kernel_mat(idx_j, idx_j);
        if (doing_sparse_kernel_mat)
            hessian = hessian - (2*kernel_mat(max([idx_i, idx_j]), min([idx_i, idx_j])));
        else
            hessian = hessian - (2*kernel_mat(idx_i, idx_j));
        end
    end

    lambda = min([(gradient_diff/hessian), (box_constraint-alpha_i), (box_constraint-alpha_j)]);   
    sv_alphas(sv_idx_i) = sv_alphas(sv_idx_i) + lambda;
    sv_alphas(sv_idx_j) = sv_alphas(sv_idx_j) + lambda;

    if (~doing_sv_preallocation)
        n_sv = length(sv_idxs);
    end
    for s = 1:n_sv
        if (doing_pt_pairwise_kernel_mat)
            if (doing_kernel_cache)
                %sv_gradients(s) = sv_gradients(s) - (lambda * (kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s)) - kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s))));

                %kv_is           = kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                %kv_js           = kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));

                %if (doing_kernel_cache_check)
                %    idx_to_row_i = kernel_cache.idx_to_row(idx_i);
                %    idx_to_row_s = kernel_cache.idx_to_row(sv_idxs(s));
                %    if ( (idx_to_row_i > 0) && (idx_to_row_s > 0) )
                %        if (idx_to_row_i > idx_to_row_s)
                %            kv_is = kernel_cache.matrix(idx_to_row_i, idx_to_row_s);
                %        else
                %            kv_is = kernel_cache.matrix(idx_to_row_s, idx_to_row_i);
                %        end
                %    else
                %        kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                %        kv_is = kernel_value;
                %    end
                %    idx_to_row_j = kernel_cache.idx_to_row(idx_j);
                %    if ( (idx_to_row_j > 0) && (idx_to_row_s > 0) )
                %        if (idx_to_row_j > idx_to_row_s)
                %            kv_js = kernel_cache.matrix(idx_to_row_j, idx_to_row_s);
                %        else
                %            kv_js = kernel_cache.matrix(idx_to_row_s, idx_to_row_j);
                %        end
                %    else             
                %        kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));
                %        kv_js = kernel_value;
                %    end
                %else
                    kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                    kv_is = kernel_value;
                    kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));
                    kv_js = kernel_value;
                %end

                sv_gradients(s) = sv_gradients(s) - (lambda * (kv_is - kv_js));
            else
                sv_gradients(s) = sv_gradients(s) - (lambda * (kernel_mat_pairwise_fn(idx_i, sv_idxs(s)) - kernel_mat_pairwise_fn(idx_j, sv_idxs(s))));
            end
        else
            if (doing_sparse_kernel_mat)
                sv_gradients(s) = sv_gradients(s) - (lambda * (kernel_mat(max([idx_i, sv_idxs(s)]), min([idx_i, sv_idxs(s)])) - kernel_mat(max([idx_j, sv_idxs(s)]), min([idx_j, sv_idxs(s)]))));
            else
                sv_gradients(s) = sv_gradients(s) - (lambda * (kernel_mat(idx_i, sv_idxs(s)) - kernel_mat(idx_j, sv_idxs(s))));
            end
        end
    end

    % Remove redundant SVs
    [gradient_i, sv_idx_i, idx_i, alpha_i] = find_max_gradient_pos_class(); 
    [gradient_j, sv_idx_j, idx_j, alpha_j] = find_min_gradient_neg_class(); 

    if (doing_sv_preallocation)
        si = 1;
        while (si <= n_sv)
            if (sv_alphas(si) == 0)
                if ( (sv_labels(si) > 0) & (sv_gradients(si) <= gradient_j) )

                    if (doing_kernel_cache)
                        kernel_cache_del_row(sv_idxs(si));
                    end

                    sv_idxs(si:(n_sv-1))      = sv_idxs((si+1):n_sv);
                    sv_alphas(si:(n_sv-1))    = sv_alphas((si+1):n_sv);
                    sv_labels(si:(n_sv-1))    = sv_labels((si+1):n_sv);
                    sv_gradients(si:(n_sv-1)) = sv_gradients((si+1):n_sv);

                    sv_idxs(n_sv)      = 0;
                    sv_labels(n_sv)    = 0;
                    sv_alphas(n_sv)    = 0;
                    sv_gradients(n_sv) = 0;

                    n_sv = n_sv - 1;

                    continue;                

                elseif (sv_gradients(si) >= gradient_i)

                    if (doing_kernel_cache)
                        kernel_cache_del_row(sv_idxs(si));
                    end

                    sv_idxs(si:(n_sv-1))      = sv_idxs((si+1):n_sv);
                    sv_alphas(si:(n_sv-1))    = sv_alphas((si+1):n_sv);
                    sv_labels(si:(n_sv-1))    = sv_labels((si+1):n_sv);
                    sv_gradients(si:(n_sv-1)) = sv_gradients((si+1):n_sv);

                    sv_idxs(n_sv)      = 0;
                    sv_labels(n_sv)    = 0;
                    sv_alphas(n_sv)    = 0;
                    sv_gradients(n_sv) = 0;

                    n_sv = n_sv - 1;

                    continue;

                end
            end

            si = si + 1;
        end
    else
        non_sv_idxs    = [];
        true_sv_idxs   = [];
        %if (doing_sv_preallocation)
        %    zero_sv_idxs   = find(sv_alphas(1:n_sv) == 0);
        %    true_sv_idxs   = find(sv_alphas(1:n_sv) > 0);
        %else
            zero_sv_idxs   = find(sv_alphas == 0);
            true_sv_idxs   = find(sv_alphas > 0);
        %end
        n_zero_sv_idxs = length(zero_sv_idxs);
        for zsi = 1:n_zero_sv_idxs
            % Debugging.
            zs = zero_sv_idxs(zsi);
            if (sv_labels(zs) > 0)
                if (sv_gradients(zs) <= gradient_j)
                    non_sv_idxs = [non_sv_idxs; zs];

                    if (doing_kernel_cache)
                        kernel_cache_del_row(sv_idxs(zs));
                    end
                else
                    true_sv_idxs = [true_sv_idxs; zs];

                    if (doing_kernel_cache)
                        kernel_cache_add_row_pairwise(sv_idxs(zs));
                    end
                end 
            else
                if (sv_gradients(zs) >= gradient_i)
                    non_sv_idxs = [non_sv_idxs; zs];

                    if (doing_kernel_cache)
                        kernel_cache_del_row(sv_idxs(zs));
                    end

                else
                    true_sv_idxs = [true_sv_idxs; zs];

                    if (doing_kernel_cache)
                        kernel_cache_add_row_pairwise(sv_idxs(zs));
                    end
                end 
            end
        end
        
        if(~isempty(non_sv_idxs))
            %true_sv_idxs = [];
            %for s = 1:n_sv            
            %    if (isempty(find(non_sv_idxs == s)))
            %        true_sv_idxs = [true_sv_idxs; s];  
            %    end
            %end

            %if (doing_sv_preallocation)
            %    n_sv                 = length(true_sv_idxs);
            %
            %    sv_idxs(1:n_sv)      = sv_idxs(true_sv_idxs);
            %    sv_labels(1:n_sv)    = sv_labels(true_sv_idxs);
            %    sv_alphas(1:n_sv)    = sv_alphas(true_sv_idxs);
            %    sv_gradients(1:n_sv) = sv_gradients(true_sv_idxs);            
            %
            %    sv_idxs(n_sv+1:end)      = 0;
            %    sv_labels(n_sv+1:end)    = 0;
            %    sv_alphas(n_sv+1:end)    = 0;
            %    sv_gradients(n_sv+1:end) = 0;
            %else
                sv_idxs      = sv_idxs(true_sv_idxs);
                sv_labels    = sv_labels(true_sv_idxs);
                sv_alphas    = sv_alphas(true_sv_idxs);
                sv_gradients = sv_gradients(true_sv_idxs);
            %end

            % Debugging
            %sv_idxs
            %sv_labels
            %sv_alphas
            %sv_gradients
        end

    end

    % Classification:
    %bias  = 0.5*(gradient_i + gradient_j);
    % Ranking:
    bias = sum(sv_alphas .* sv_labels);

    delta = gradient_i - gradient_j;
end

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [] = lasvm_finish_reprocess(tau)

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;
global bias;
global delta;

global doing_sv_preallocation;

while (delta > tau)
    % Debugging.
    %disp('Finishing Reprocess.............')
    lasvm_reprocess(tau);
end

if (doing_sv_preallocation)
    non_zero_sv_idxs = find(sv_alphas(1:n_sv) > 0);

    n_sv = length(non_zero_sv_idxs);

    sv_idxs(1:n_sv)      = sv_idxs(non_zero_sv_idxs);
    sv_labels(1:n_sv)    = sv_labels(non_zero_sv_idxs);
    sv_alphas(1:n_sv)    = sv_alphas(non_zero_sv_idxs);
    sv_gradients(1:n_sv) = sv_gradients(non_zero_sv_idxs);

    sv_idxs(n_sv+1:end)      = 0;
    sv_labels(n_sv+1:end)    = 0;
    sv_alphas(n_sv+1:end)    = 0;
    sv_gradients(n_sv+1:end) = 0;
else
    non_zero_sv_idxs = find(sv_alphas > 0);
    sv_idxs          = sv_idxs(non_zero_sv_idxs);
    sv_labels        = sv_labels(non_zero_sv_idxs);
    sv_alphas        = sv_alphas(non_zero_sv_idxs);
    sv_gradients     = sv_gradients(non_zero_sv_idxs);

    n_sv = length(sv_idxs);
end

% Debugging.
%sv_idxs
%sv_labels
%sv_alphas
%sv_gradients
bias;
delta;

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [gradient_i, sv_idx_i, idx_i, alpha_i] = find_max_gradient_pos_class()

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;
global bias;
global delta;

global box_constraint;

global doing_sv_preallocation;

if (doing_sv_preallocation)
    pos_sv_idxs = find((sv_labels(1:n_sv) > 0) & (sv_alphas(1:n_sv) < box_constraint));
else
    pos_sv_idxs = find((sv_labels > 0) & (sv_alphas < box_constraint));
end

if (~isempty(pos_sv_idxs))
    pos_sv_gradients      = sv_gradients(pos_sv_idxs);
    [gradient_i, max_idx] = max(pos_sv_gradients);         
    sv_idx_i              = pos_sv_idxs(max_idx);
    idx_i                 = sv_idxs(sv_idx_i); 
    alpha_i               = sv_alphas(sv_idx_i);
else
    gradient_i = [];
    sv_idx_i   = [];
    idx_i      = [];
    alpha_i    = [];
end

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [gradient_i, sv_idx_i, idx_i, alpha_i] = find_min_gradient_pos_class()

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;
global bias;
global delta;

global box_constraint;

global doing_sv_preallocation;

if (doing_sv_preallocation)
    pos_sv_idxs = find(sv_labels(1:n_sv) > 0);
else
    %pos_sv_idxs = find((sv_labels > 0) & (sv_alphas < box_constraint))
    pos_sv_idxs = find(sv_labels > 0);
end

if (~isempty(pos_sv_idxs))
    pos_sv_gradients      = sv_gradients(pos_sv_idxs);
    [gradient_i, min_idx] = min(pos_sv_gradients);         
    sv_idx_i              = pos_sv_idxs(min_idx);
    idx_i                 = sv_idxs(sv_idx_i); 
    alpha_i               = sv_alphas(sv_idx_i);
else
    gradient_i = [];
    sv_idx_i   = [];
    idx_i      = [];
    alpha_i    = [];
end

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [gradient_j, sv_idx_j, idx_j, alpha_j] = find_max_gradient_neg_class()

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;
global bias;
global delta;

global box_constraint;

global doing_sv_preallocation;

if (doing_sv_preallocation)
    neg_sv_idxs = find(sv_labels(1:n_sv) < 0);
else
    %neg_sv_idxs = find((sv_labels < 0) & (sv_alphas < box_constraint))
    neg_sv_idxs = find(sv_labels < 0);
end

if (~isempty(neg_sv_idxs))
    neg_sv_gradients      = sv_gradients(neg_sv_idxs);
    [gradient_j, max_idx] = max(neg_sv_gradients);         
    sv_idx_j              = neg_sv_idxs(max_idx);
    idx_j                 = sv_idxs(sv_idx_j);        
    alpha_j               = sv_alphas(sv_idx_j);
else
    gradient_j = [];
    sv_idx_j   = [];
    idx_j      = [];
    alpha_j    = [];
end

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [gradient_j, sv_idx_j, idx_j, alpha_j] = find_min_gradient_neg_class()

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;
global bias;
global delta;

global box_constraint;

global doing_sv_preallocation;

if (doing_sv_preallocation)
    neg_sv_idxs = find((sv_labels(1:n_sv) < 0) & (sv_alphas(1:n_sv) < box_constraint));
else
    neg_sv_idxs = find((sv_labels < 0) & (sv_alphas < box_constraint));
end

if (~isempty(neg_sv_idxs))
    neg_sv_gradients      = sv_gradients(neg_sv_idxs);
    [gradient_j, min_idx] = min(neg_sv_gradients);         
    sv_idx_j              = neg_sv_idxs(min_idx);
    idx_j                 = sv_idxs(sv_idx_j);        
    alpha_j               = sv_alphas(sv_idx_j);
else
    gradient_j = [];
    sv_idx_j   = [];
    idx_j      = [];
    alpha_j    = [];
end

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [] = eliminate_svs_single_gradient(max_num_sv, tau, reoptimize)

% Global.
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;

global kernel_mat;
global kernel_cache;
global box_constraint;
global delta;

global doing_kernel_cache;
global doing_pt_pairwise_kernel_mat;
global doing_sparse_kernel_mat;
global doing_sq_error;

global orderings_pt_idxs;

n_sv = length(sv_idxs);
self_kernel_vec = zeros(n_sv, 1);
while (n_sv > max_num_sv)

    n_sv = length(sv_idxs);
    self_kernel_vec = zeros(n_sv, 1);
    for s = 1:n_sv
        if (doing_pt_pairwise_kernel_mat)
            if (doing_kernel_cache)
                self_kernel_vec(s) = kernel_mat_pairwise_cache_fn(sv_idxs(s), sv_idxs(s));
            else
                self_kernel_vec(s) = kernel_mat_pairwise_fn(sv_idxs(s), sv_idxs(s));
            end
        else
            self_kernel_vec(s) = kernel_mat(sv_idxs(s), sv_idxs(s));
        end
    end
    delta_objective = (0.5 * self_kernel_vec .* (sv_alphas.^2)) + ((-1*sv_labels.*sv_gradients) .* (-1*sv_alphas));

    [min_delta_obj, min_sv_idx] = min(delta_objective);
    min_idx                     = sv_idxs(min_sv_idx);
    min_label                   = sv_labels(min_sv_idx); 
    min_alpha                   = sv_alphas(min_sv_idx);
    retain_sv_idxs              = find(delta_objective ~= min_delta_obj);

    other_class_sv_idxs   = find(sv_labels ~= min_label);
    max_alpha_other_class = max(sv_alphas(other_class_sv_idxs));
    n_other_class         = length(other_class_sv_idxs);
    
    if (max_alpha_other_class >= min_alpha)
        % Debugging.
        %disp('Eliminating a SV.............')

        for so = 1:n_other_class
            soi = other_class_sv_idxs(so);
            if (sv_alphas(soi) >= max_alpha_other_class)
                sv_alphas(soi) = sv_alphas(soi) - min_alpha;    
                sv_idx_soi     = soi;
                idx_soi        = sv_idxs(soi);
                break;
            end
        end

        sv_idxs         = sv_idxs(retain_sv_idxs);
        sv_labels       = sv_labels(retain_sv_idxs);
        sv_alphas       = sv_alphas(retain_sv_idxs);
        sv_gradients    = sv_gradients(retain_sv_idxs);    
        self_kernel_vec = self_kernel_vec(retain_sv_idxs);    

        n_sv = length(sv_idxs);
        for s = 1:n_sv
            idx_s   = sv_idxs(s);
            label_s = sv_labels(s);
            if (doing_pt_pairwise_kernel_mat)
                if(doing_kernel_cache)
                    sv_gradients(s) = sv_gradients(s) + (min_alpha * ((min_label * kernel_mat_pairwise_cache_fn(min_idx, idx_s)) - (min_label * kernel_mat_pairwise_cache_fn(idx_soi, idx_s))));
                else
                    sv_gradients(s) = sv_gradients(s) + (min_alpha * ((min_label * kernel_mat_pairwise_fn(min_idx, idx_s)) - (min_label * kernel_mat_pairwise_fn(idx_soi, idx_s))));
                end
            else
                if (doing_sparse_kernel_mat)
                    sv_gradients(s) = sv_gradients(s) + (min_alpha * ((min_label * kernel_mat(max([min_idx, idx_s]), min([min_idx, idx_s]))) - ...
                                                                      (min_label * kernel_mat(max([idx_soi, idx_s]), min([idx_soi, idx_s])))));
                else
                    sv_gradients(s) = sv_gradients(s) + (min_alpha * ((min_label * kernel_mat(min_idx, idx_s)) - (min_label * kernel_mat(idx_soi, idx_s))));
                end
            end

        end

        % Reoptimize alphas.
        if (reoptimize == 1)
            [gradient_i, sv_idx_i, idx_i, alpha_i] = find_max_gradient_pos_class();    
	    [gradient_j, sv_idx_j, idx_j, alpha_j] = find_min_gradient_neg_class();    
            if ( (isempty(idx_i)) | (isempty(idx_j)) )
	        delta = 0;
	    else
	        gradient_diff = gradient_i - gradient_j;
	        delta         = gradient_diff;
            end 
            lasvm_finish_reprocess(tau);
        end
    else
        return;
    end
end

% Debugging.
%sv_idxs
%sv_labels
%sv_alphas
%sv_gradients

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [] = eliminate_svs_single_gradient_with_redistribution(max_num_sv, tau, reoptimize)

% Global.
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;

global kernel_mat;
global kernel_cache;
global box_constraint;
global delta;

global doing_kernel_cache;
global doing_pt_pairwise_kernel_mat;
global doing_sparse_kernel_mat;
global doing_sq_error;

global orderings_pt_idxs;

n_sv = length(sv_idxs);
while (n_sv > max_num_sv)

    self_kernel_vec = zeros(n_sv, 1);
    for s = 1:n_sv
        if (doing_pt_pairwise_kernel_mat)
            if (doing_kernel_cache)
                self_kernel_vec(s) = kernel_mat_pairwise_cache_fn(sv_idxs(s), sv_idxs(s));
            else
                self_kernel_vec(s) = kernel_mat_pairwise_fn(sv_idxs(s), sv_idxs(s));
            end
        else
            self_kernel_vec(s) = kernel_mat(sv_idxs(s), sv_idxs(s));
        end
    end

    % Debugging.
    %size(self_kernel_vec)
    %size(sv_alphas)
    %size(sv_labels)
    %size(sv_gradients)
 
    delta_objective = (0.5 * self_kernel_vec .* (sv_alphas.^2)) + ((-1*sv_labels.*sv_gradients) .* (-1*sv_alphas));

    [min_delta_obj, min_sv_idx] = min(delta_objective);
    min_idx                     = sv_idxs(min_sv_idx);
    min_label                   = sv_labels(min_sv_idx); 
    min_alpha                   = sv_alphas(min_sv_idx);
    retain_sv_idxs              = find(delta_objective ~= min_delta_obj);
    
    rem_pos_sv_idxs = find( (sv_labels > 0) & (sv_idxs ~= min_idx) );
    rem_neg_sv_idxs = find( (sv_labels < 0) & (sv_idxs ~= min_idx) );

    room_check = 0;
    while ( (room_check == 0) & (~isempty(rem_pos_sv_idxs)) & (~isempty(rem_neg_sv_idxs)) )
	    n_rem_pos_sv = length(rem_pos_sv_idxs);
	    n_rem_neg_sv = length(rem_neg_sv_idxs);

	    sum_rem_pos_alphas = sum(sv_alphas(rem_pos_sv_idxs));
	    sum_rem_neg_alphas = sum(sv_alphas(rem_neg_sv_idxs));

	    rem_pos_alphas = sv_alphas(rem_pos_sv_idxs);
	    rem_neg_alphas = sv_alphas(rem_neg_sv_idxs);

	    epsilon = (sum_rem_neg_alphas - sum_rem_pos_alphas) / (n_rem_pos_sv + n_rem_neg_sv);

	    if (epsilon > 0)
		[max_rem_pos_alpha, max_rem_pos_alpha_idx] = max(rem_pos_alphas);
		[min_rem_neg_alpha, min_rem_neg_alpha_idx] = min(rem_neg_alphas);

		[max_room, max_room_idx] = min([(box_constraint - max_rem_pos_alpha), min_rem_neg_alpha]);
		if (epsilon <= max_room)
		    room_check = 1;
		else
		    room_check = 0;
		end       
	    else
		[min_rem_pos_alpha, min_rem_pos_alpha_idx] = min(rem_pos_alphas);
		[max_rem_neg_alpha, max_rem_neg_alpha_idx] = max(rem_neg_alphas);

		[max_room, max_room_idx] = min([min_rem_pos_alpha, (box_constraint - max_rem_neg_alpha)]);
		if (abs(epsilon) <= max_room)
		    room_check = 1;
		else
		    room_check = 0;
		end       
	    end

            % Debugging.
            %room_check

	    if (room_check == 1)
		sv_alphas(rem_pos_sv_idxs) = sv_alphas(rem_pos_sv_idxs) + epsilon;
		sv_alphas(rem_neg_sv_idxs) = sv_alphas(rem_neg_sv_idxs) - epsilon;		  

		n_sv = length(sv_idxs);
		for s = 1:n_sv
		    idx_s   = sv_idxs(s);
		    label_s = sv_labels(s);
                    if (doing_pt_pairwise_kernel_mat)
                        if (doing_kernel_cache)
		                    sv_gradients(s) = sv_gradients(s) + (min_alpha * min_label * kernel_mat_pairwise_cache_fn(min_idx, idx_s));
                        else
		                    sv_gradients(s) = sv_gradients(s) + (min_alpha * min_label * kernel_mat_pairwise_fn(min_idx, idx_s));
                        end
                    else
                        if (doing_sparse_kernel_mat)
        		            sv_gradients(s) = sv_gradients(s) + (min_alpha * min_label * kernel_mat(max([min_idx, idx_s]), min([min_idx, idx_s])));
                        else
		                    sv_gradients(s) = sv_gradients(s) + (min_alpha * min_label * kernel_mat(min_idx, idx_s));
                        end
                    end
		    for s_rp = 1:n_rem_pos_sv
			            rp_idx          = sv_idxs(rem_pos_sv_idxs(s_rp));
                        if (doing_pt_pairwise_kernel_mat)
                            if (doing_kernel_cache)
            			        sv_gradients(s) = sv_gradients(s) - (epsilon * kernel_mat_pairwise_cache_fn(rp_idx, idx_s));
                            else
            			        sv_gradients(s) = sv_gradients(s) - (epsilon * kernel_mat_pairwise_fn(rp_idx, idx_s));
                            end
                        else
                            if (doing_sparse_kernel_mat)
			                    sv_gradients(s) = sv_gradients(s) - (epsilon * kernel_mat(max([rp_idx, idx_s]), min([rp_idx, idx_s])));
                            else
            			        sv_gradients(s) = sv_gradients(s) - (epsilon * kernel_mat(rp_idx, idx_s));
                            end
                        end
		    end
		    for s_rn = 1:n_rem_neg_sv
    			rn_idx          = sv_idxs(rem_neg_sv_idxs(s_rn));
                        if (doing_pt_pairwise_kernel_mat)
                            if (doing_kernel_cache)
	            		        sv_gradients(s) = sv_gradients(s) - (epsilon * kernel_mat_pairwise_cache_fn(rn_idx, idx_s));
                            else
	            		        sv_gradients(s) = sv_gradients(s) - (epsilon * kernel_mat_pairwise_fn(rn_idx, idx_s));
                            end
                        else
                            if (doing_sparse_kernel_mat)
			                    sv_gradients(s) = sv_gradients(s) - (epsilon * kernel_mat(max([rn_idx, idx_s]), min([rn_idx, idx_s])));
                            else
            			        sv_gradients(s) = sv_gradients(s) - (epsilon * kernel_mat(rn_idx, idx_s));
                            end
                        end
		    end
		end

                sv_idxs         = sv_idxs(retain_sv_idxs);
		sv_labels       = sv_labels(retain_sv_idxs);
		sv_alphas       = sv_alphas(retain_sv_idxs);
		sv_gradients    = sv_gradients(retain_sv_idxs);    
                self_kernel_vec = self_kernel_vec(retain_sv_idxs);  

                % Reoptimize alphas.
                if (reoptimize == 1)
                    [gradient_i, sv_idx_i, idx_i, alpha_i] = find_max_gradient_pos_class();    
        	    [gradient_j, sv_idx_j, idx_j, alpha_j] = find_min_gradient_neg_class();    
		    if ( (isempty(idx_i)) | (isempty(idx_j)) )
		        delta = 0;
		    else
		        gradient_diff = gradient_i - gradient_j;
		        delta         = gradient_diff;
        	    end 
                    lasvm_finish_reprocess(tau);
                end
	    else
		if (epsilon > 0)
		    if (max_room_idx == 1)
			rem_pos_sv_idxs = rem_pos_sv_idxs(find(rem_pos_sv_idxs ~= rem_pos_sv_idxs(max_rem_pos_alpha_idx)));            
		    else
			rem_neg_sv_idxs = rem_neg_sv_idxs(find(rem_neg_sv_idxs ~= rem_neg_sv_idxs(min_rem_neg_alpha_idx)));                            
		    end
		else
		    if (max_room_idx == 1)
			rem_pos_sv_idxs = rem_pos_sv_idxs(find(rem_pos_sv_idxs ~= rem_pos_sv_idxs(min_rem_pos_alpha_idx)));            
		    else
			rem_neg_sv_idxs = rem_neg_sv_idxs(find(rem_neg_sv_idxs ~= rem_neg_sv_idxs(max_rem_neg_alpha_idx)));                            
		    end
		end
	    end

	    n_sv = length(sv_idxs); 
    end

    if ( isempty(rem_pos_sv_idxs) | isempty(rem_neg_sv_idxs) )
        break;
    end   
end

% Debugging.
%sv_idxs
%sv_labels
%sv_alphas
%sv_gradients

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [] = eliminate_svs_pairwise_gradient(max_num_sv, tau, reoptimize)

% Global.
global n_sv;
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;

global kernel_mat;
global kernel_cache;
global box_constraint;
global delta;

global doing_kernel_cache;
global doing_kernel_cache_check;
global doing_pt_pairwise_kernel_mat;
global doing_sparse_kernel_mat;
global doing_sq_error;
global doing_sv_preallocation;

global orderings_pt_idxs;

% Pre-allocate variables to hold kernel values.
global kernel_value;

kernel_value = 0;
kv_is        = 0;
kv_js        = 0;

if (~doing_sv_preallocation)
    n_sv = length(sv_idxs);
end

%self_kernel_vec = zeros(n_sv, 1);
while (n_sv > max_num_sv)

    [gradient_i, sv_idx_i, idx_i, alpha_i] = find_min_gradient_pos_class();    
    if (isempty(idx_i))
        return;
    end

    [gradient_j, sv_idx_j, idx_j, alpha_j] = find_max_gradient_neg_class();    
    if (isempty(idx_j))
        return;
    end

    % Debugging.
    %(gradient_i - gradient_j)

    if (alpha_i < alpha_j)
        min_alpha = alpha_i;
    else
        min_alpha = alpha_j;
    end    

    sv_alphas(sv_idx_i) = sv_alphas(sv_idx_i) - min_alpha;
    sv_alphas(sv_idx_j) = sv_alphas(sv_idx_j) - min_alpha;

    for s = 1:n_sv
        idx_s   = sv_idxs(s);
        label_s = sv_labels(s);
        if (doing_pt_pairwise_kernel_mat)
            if (doing_kernel_cache)
                %sv_gradients(s) = sv_gradients(s) + (min_alpha * (kernel_mat_pairwise_cache_fn(idx_i, idx_s) - kernel_mat_pairwise_cache_fn(idx_j, idx_s)));

                %kv_is           = kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                %kv_js           = kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));

                %if (doing_kernel_cache_check)
                %    idx_to_row_i = kernel_cache.idx_to_row(idx_i);
                %    idx_to_row_s = kernel_cache.idx_to_row(sv_idxs(s));
                %    if ( (idx_to_row_i > 0) && (idx_to_row_s > 0) )
                %        if (idx_to_row_i > idx_to_row_s)
                %            kv_is = kernel_cache.matrix(idx_to_row_i, idx_to_row_s);
                %        else
                %            kv_is = kernel_cache.matrix(idx_to_row_s, idx_to_row_i);
                %        end
                %    else
                %        kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                %        kv_is = kernel_value;
                %    end
                %    idx_to_row_j = kernel_cache.idx_to_row(idx_j);
                %    if ( (idx_to_row_j > 0) && (idx_to_row_s > 0) )
                %        if (idx_to_row_j > idx_to_row_s)
                %            kv_js = kernel_cache.matrix(idx_to_row_j, idx_to_row_s);
                %        else
                %            kv_js = kernel_cache.matrix(idx_to_row_s, idx_to_row_j);
                %        end
                %    else                             
                %        kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));
                %        kv_js = kernel_value;
                %    end
                %else
                    kernel_mat_pairwise_cache_fn(idx_i, sv_idxs(s));
                    kv_is = kernel_value;
                    kernel_mat_pairwise_cache_fn(idx_j, sv_idxs(s));
                    kv_js = kernel_value;
                %end

                sv_gradients(s) = sv_gradients(s) + (min_alpha * (kv_is - kv_js));
            else
                sv_gradients(s) = sv_gradients(s) + (min_alpha * (kernel_mat_pairwise_fn(idx_i, idx_s) - kernel_mat_pairwise_fn(idx_j, idx_s)));
            end
        else
            if (doing_sparse_kernel_mat)
                sv_gradients(s) = sv_gradients(s) + (min_alpha * (kernel_mat(max([idx_i, idx_s]), min([idx_i, idx_s])) - kernel_mat(max([idx_j, idx_s]), min([idx_j, idx_s]))));
            else
                sv_gradients(s) = sv_gradients(s) + (min_alpha * (kernel_mat(idx_i, idx_s) - kernel_mat(idx_j, idx_s)));
            end
        end
    end

    if (doing_sv_preallocation)
        non_zero_sv_idxs = find(sv_alphas(1:n_sv) > 0);

        n_sv = length(non_zero_sv_idxs);

        sv_idxs(1:n_sv)      = sv_idxs(non_zero_sv_idxs);
        sv_labels(1:n_sv)    = sv_labels(non_zero_sv_idxs);
        sv_alphas(1:n_sv)    = sv_alphas(non_zero_sv_idxs);
        sv_gradients(1:n_sv) = sv_gradients(non_zero_sv_idxs);

        sv_idxs(n_sv+1:end)      = 0;
        sv_labels(n_sv+1:end)    = 0;
        sv_alphas(n_sv+1:end)    = 0;
        sv_gradients(n_sv+1:end) = 0;
    else
        non_zero_sv_idxs = find(sv_alphas > 0);
        sv_idxs          = sv_idxs(non_zero_sv_idxs);
        sv_labels        = sv_labels(non_zero_sv_idxs);
        sv_alphas        = sv_alphas(non_zero_sv_idxs);
        sv_gradients     = sv_gradients(non_zero_sv_idxs);

        n_sv = length(sv_idxs);
    end

    % Reoptimize alphas.
    if (reoptimize == 1)
        [gradient_i, sv_idx_i, idx_i, alpha_i] = find_max_gradient_pos_class();   
        [gradient_j, sv_idx_j, idx_j, alpha_j] = find_min_gradient_neg_class();    
        if ( (isempty(idx_i)) | (isempty(idx_j)) )
    	    delta = 0;
        else
    	    gradient_diff = gradient_i - gradient_j;
    	    delta         = gradient_diff;
        end 
        lasvm_finish_reprocess(tau);

        if (~doing_sv_preallocation)
            n_sv = length(sv_idxs);
        end
    end

end

% Debugging.
%sv_idxs
%sv_labels
%sv_alphas
%sv_gradients

return;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function[] = plot_decision_boundary(kernel, orderings, use_sigmoid)

% Global.
global sv_idxs;
global sv_labels;
global sv_alphas;
global sv_gradients;

global pts;
global labels;
global bias;

global gauss_sigma;
global poly_degree;
global poly_bias;

n_pts = size(pts, 2);

% Plot the decision boundary.
figure, hold on;

xx   = -2.0:0.1:2.0;
yy   = -2.0:0.1:2.0;
n_xx = length(xx);
n_yy = length(yy);

this_prob1 = zeros(n_yy, n_xx);

for ii = 1:n_xx
    for jj = 1:n_yy
        w_phi = 0;
        n_sv = length(sv_idxs);
        for j = 1:n_sv
	    idx_j   = sv_idxs(j);
	    y_j     = sv_labels(j);
	    alpha_j = sv_alphas(j);

            x_hi  = pts(:, find(orderings(idx_j, :) == 1));
            x_low = pts(:, find(orderings(idx_j, :) == -1));

            if (~isempty(strfind(kernel, 'gauss')))
	        w_phi = w_phi + (alpha_j * y_j * (  feval(kernel, x_hi, [xx(ii); yy(jj)], gauss_sigma) ...
                                                  - feval(kernel, x_low, [xx(ii); yy(jj)], gauss_sigma)));
            elseif (~isempty(strfind(kernel, 'poly')))
	        w_phi = w_phi + (alpha_j * y_j * (  feval(kernel, x_hi, [xx(ii); yy(jj)], poly_degree, poly_bias) ... 
                                                  - feval(kernel, x_low, [xx(ii); yy(jj)], poly_degree, poly_bias)));
            else
	        w_phi = w_phi + (alpha_j * y_j * ( feval(kernel, x_hi, [xx(ii); yy(jj)]) ...
                                                 - feval(kernel, x_low, [xx(ii); yy(jj)])));                
            end
        end
        wphi = w_phi + bias;

        if (use_sigmoid == 1)
	    this_prob1(jj, ii) = 1/(1+exp(-w_phi));
        else
	    this_prob1(jj, ii) = sign(w_phi);
        end
    end
end

% Debugging.
%this_prob1

%for j = 1:n_pts
%    if (labels(j) > 0)
%        if (~isempty(find(sv_idxs == j)))
%	    plot(pts(1,j), pts(2,j), 'gs', 'MarkerEdgeColor', 'r', 'MarkerFaceColor', 'g', 'MarkerSize', ceil(10*1.0));
%        else
%	    plot(pts(1,j), pts(2,j), 'gs', 'MarkerEdgeColor', 'k', 'MarkerFaceColor', 'g', 'MarkerSize', ceil(10*1.0));
%        end
%    else
%        if (~isempty(find(sv_idxs == j)))
%	    plot(pts(1,j), pts(2,j), 'bd', 'MarkerEdgeColor', 'r', 'MarkerFaceColor', 'b', 'MarkerSize', ceil(10*1.0));
%        else
%	    plot(pts(1,j), pts(2,j), 'bd', 'MarkerEdgeColor', 'k', 'MarkerFaceColor', 'b', 'MarkerSize', ceil(10*1.0));
%        end
%    end
%end
for j = 1:n_pts
    plot(pts(1,j), pts(2,j), 'gs', 'MarkerEdgeColor', 'r', 'MarkerFaceColor', 'g', 'MarkerSize', ceil(10*1.0));
end

pcolor(xx, yy, this_prob1);
colormap(gray);
shading interp;

min_xx = min(xx);
max_xx = max(xx);
min_yy = min(min(yy), min(pts(2, :)));
max_yy = max(max(yy), max(pts(1, :)));
axis([min_xx max_xx min_yy max_yy]);

return;

end
