% Kernel Rank SVM testing function.
%
% Function parameters:
% ====================
% test_pts: A matrix of test points. Each column of the matrix is one point and
% each row is a feature dimension.
%
% model: A structure storing the parameters of the learnt model. Please see the
% documentation of kernel_rank_lasvm_sgd_train() for the details of the
% parameters.
%
% model_mat_file: Path to a matlab mat file containing the kernel ranking SVM
% parameters. Please refer to the documentation of kernel_rank_lasvm_sgd_train()
% for the details of the parameters. If model_mat_file is empty, then the model
% parameters are read in from the previous parameter directly. Otherwise, the
% model parameters are read in by loading this file (supersedes previous
% 'model' parameter). 
%
% use_sigmoid: Outputs score converted to a probability (number between 0 and 1)
% when set to 1, otherwise outputs raw scores. The transformation to
% probability values is done by putting the raw scores through a sigmoid
% function.
%
% Function outputs:
% =================
% test_labels: Vector of prediction scores for the test points. 

function [test_labels] = kernel_ranksvm_test(test_pts, ...
                                             model, ...
                                             model_mat_file, ...
                                             use_sigmoid)

verbose = 0;

if (~isempty(model_mat_file))
    load(model_mat_file);
end

mean_pts    = model.mean_pts;
std_pts     = model.std_pts;
sv_idxs     = model.sv_idxs;
sv_pts      = model.sv_pts;
sv_alphas   = model.sv_alphas;
sv_labels   = model.sv_labels;
bias        = model.bias;
kernel      = model.kernel;
gauss_sigma = model.gauss_sigma;
poly_degree = model.poly_degree;
poly_bias   = model.poly_bias;

n_sv        = length(sv_idxs);
n_pts       = size(test_pts, 2);
n_dim       = size(test_pts, 1);
test_labels = zeros(n_pts, 1);

% Z normalization.
if (verbose > 0)
    disp('Performing unit normalization.............');
    tic
end
% In theory, the following is slow.
%test_pts = test_pts - repmat(mean_pts, [1 n_pts]);
%test_pts = test_pts ./ repmat(std_pts, [1 n_pts]);

test_pts = test_pts - mean_pts(:, ones(n_pts, 1));
test_pts = test_pts ./ std_pts(:, ones(n_pts, 1));
if (verbose > 0)
    toc
end

if (verbose > 0)
    disp('Predicting labels.............');
    tic
end
for i = 1:n_pts
    w_phi = 0;
    for j = 1:n_sv
	sv_idx_j = sv_idxs(j);        
	y_j      = sv_labels(j);
	alpha_j  = sv_alphas(j);

        %hi_idx  = find(O_(sv_idx_j, :) == 1);
        %low_idx = find(O_(sv_idx_j, :) == -1);
        %if (y_j == 1)
        %    x_hi  = train_pts(:, hi_idx);
        %    x_low = train_pts(:, low_idx);
        %else
        %    x_hi  = train_pts(:, low_idx);
        %    x_low = train_pts(:, hi_idx);
        %end

        x_hi  = sv_pts(1:n_dim, j);
        x_low = sv_pts((n_dim+1):(2*n_dim), j);

        if (~isempty(strfind(kernel, 'gauss')))
	    w_phi = w_phi + (alpha_j * y_j * (feval(kernel, x_hi, test_pts(:, i), gauss_sigma) - feval(kernel, x_low, test_pts(:, i), gauss_sigma)));
        elseif (~isempty(strfind(kernel, 'poly')))
	    w_phi = w_phi + (alpha_j * y_j * (feval(kernel, x_hi, test_pts(:, i), poly_degree, poly_bias) - feval(kernel, x_low, test_pts(:, i), poly_degree, poly_bias)));
        else
	    w_phi = w_phi + (alpha_j * y_j * (feval(kernel, x_hi, test_pts(:, i)) - feval(kernel, x_low, test_pts(:, i))));            
        end
    end
    wphi = w_phi + bias;

    if (use_sigmoid == 1)
	test_labels(i) = 1/(1+exp(-w_phi));
    else
	test_labels(i) = w_phi;
    end
end
if (verbose > 0)
    toc
end

return;

end
