% Wrapper function for kernel_rank_lasvm_sgd_train(), which does the actual
% training.
%
% Function parameters:
% ====================
% X_: Data matrix with each row corresponding to a data point and each column a
% feature dimension. 
%
% O_: A sparse matrix of point pairwise orderings with each row specifying
% ordering for one such pair. The number of columns must equal the total number
% of points in the training set. On each row, the column corresponding to the
% index (within X_) of the higher ranked point in the order is set to 1
% and the column corresponding to the index (within X_) of the lower
% ranked point is set to -1. All other columns must be set to 0.
%
% C_O: A vector of training costs for penalizing the misrankings of the ordered
% pairs specified by the corresponding rows of the O_ matrix.
%
% model_mat_file: A Matlab mat file to store the parameters of the learned
% ranking model. Please the documentation of kernel_rank_lasvm_sgd_train() for
% the details of the parameters. It can be an empty matrix if you do not want to
% save the model file to disk. 
%
% Function outputs:
% =================
% model: A structure storing the parameters of the learnt model. Please see the
% documentation of kernel_rank_lasvm_sgd_train() for the details of the
% parameters.
% 
% NOTES:
% ======
% This function declares kernel_mat as a global variable, which is also
% accessible to kernel_rank_lasvm_sgd_train(). Once the kernel matrix is
% computed in this function, it is not recomputed again by
% kernel_rank_lasvm_sgd_train(). It is helpful because kernel matrix computation
% is expensive, and can be reused for multiple runs, eg. parameter search.

function [model] = ranksvm_kernel_sgd(X_, O_, C_O, model_mat_file)

global kernel_mat

% Use a kernel cache.
do_kernel_cache              = 1;
do_kernel_cache_check        = 0;
% Compute point-pairwise kernel matrix instead of order-pairwise.
do_point_pairwise_kernel_mat = 1;
% Use sparse instead of a full kernel matrix to save memory.
% It makes sense only when the kernel matrix is symmetric.
do_sparse_kernel_mat         = 0;

% Preallocate all SV related vectors.
do_sv_preallocation          = 1;

% Random number seed.
rng(6675068);
%rand('seed', 6675068);

% To avoid passing these matrices as arguments to subfunctions.
X = X_'; A = O_;

n_O = size(A,1);
d   = size(X,1);
n   = size(X,2);

% Labels.
% Randomly assign -1 to half the labels and flip the order.
y = ones(n_O, 1);
for i = 1:n_O
    if (rand < 0.5)
        hi_idx  = find(A(i, :) == 1);
        low_idx = find(A(i, :) == -1);

        y(i) = -1;
        A(i, hi_idx)  = -1;
        A(i, low_idx) = 1;
    end
end

% Parameters.
kernel             = 'gauss_kernel';
%kernel             = 'linear_kernel';
% Gaussian kernel:
gauss_kernel_scale = 0.02;
% Polynomial kernel:
poly_degree        = 2;
poly_bias          = 1;
% LASVM:
lambda             = 1.0/max(C_O);
tau                = 0.001;
z_normalize        = 0;
n_epochs           = 4;
kernel_mat         = [];
elim_methods       = {'none', 'pairwise_gradient'};
n_elim_meth        = length(elim_methods);
em_idx             = 1;
do_squared_error   = 1;
do_budget          = 0;
max_num_sv         = ceil(1.0*n_O)
%max_num_sv         = ceil(0.1*n_O)
use_sigmoid        = 1;
compute_obj        = 0;
plot_bndry         = 0;

disp('Computing kernel matrix.............');
tic
if (do_point_pairwise_kernel_mat)
    disp('Computing point-pairwise kernel matrix.............');
    if (do_sparse_kernel_mat)
        disp('Computing sparse kernel matrix.............');
        kernel_mat = sparse([], [], [], n, n, (n*(n+1)/2));
        for i = 1:n
            for j = 1:i        
                if (~isempty(strfind(kernel, 'gauss')))
                    kernel_mat(i, j) = feval(kernel, X(:, i), X(:, j), gauss_kernel_scale);
                elseif (~isempty(strfind(kernel, 'poly')))
                    kernel_mat(i, j) = feval(kernel, X(:, i), X(:, j), poly_degree, poly_bias);
                else
                   kernel_mat(i, j) = feval(kernel, X(:, i), X(:, j));
                end
            end
        end
    else
        disp('Computing full kernel matrix.............');
        kernel_mat = zeros(n, n);
        for i = 1:n
            for j = 1:n        
                if (~isempty(strfind(kernel, 'gauss')))
                    kernel_mat(i, j) = feval(kernel, X(:, i), X(:, j), gauss_kernel_scale);
                elseif (~isempty(strfind(kernel, 'poly')))
                    kernel_mat(i, j) = feval(kernel, X(:, i), X(:, j), poly_degree, poly_bias);
                else
                   kernel_mat(i, j) = feval(kernel, X(:, i), X(:, j));
                end
            end
        end
    end
else
    disp('Computing rank-pairwise kernel matrix.............');
    if (do_sparse_kernel_mat)
        disp('Computing sparse kernel matrix.............');
        kernel_mat = sparse([], [], [], n_O, n_O, (n_O*(n_O+1)/2));
        for o_1 = 1:n_O
            for o_2 = 1:o_1
                i = find(A(o_1,:) == 1);
                j = find(A(o_1,:) == -1);
                k = find(A(o_2,:) == 1);
                l = find(A(o_2,:) == -1);

                if (~isempty(strfind(kernel, 'gauss')))
	            kernel_mat(o_1, o_2) = (1 + feval(kernel, X(:, i), X(:, k), gauss_kernel_scale) ...
                                              - feval(kernel, X(:, i), X(:, l), gauss_kernel_scale) ...
                                              - feval(kernel, X(:, j), X(:, k), gauss_kernel_scale) ...
                                              + feval(kernel, X(:, j), X(:, l), gauss_kernel_scale)    );
                elseif (~isempty(strfind(kernel, 'poly')))
	            kernel_mat(o_1, o_2) = (1 + feval(kernel, X(:, i), X(:, k), poly_degree, poly_bias) ...
                                              - feval(kernel, X(:, i), X(:, l), poly_degree, poly_bias) ...
                                              - feval(kernel, X(:, j), X(:, k), poly_degree, poly_bias) ...
                                              + feval(kernel, X(:, j), X(:, l), poly_degree, poly_bias)    );
                else
	            kernel_mat(o_1, o_2) = (1 + feval(kernel, X(:, i), X(:, k)) ...
                                              - feval(kernel, X(:, i), X(:, l)) ...
                                              - feval(kernel, X(:, j), X(:, k)) ...
                                              + feval(kernel, X(:, j), X(:, l))    );
                end
       
                if (do_squared_error)
                    if (o_1 == o_2)
                        kernel_mat(o_1, o_2) = kernel_mat(o_1, o_2) + lambda;
                    end
                end
            end
        end
    else
        disp('Computing full kernel matrix.............');
        kernel_mat = zeros(n_O, n_O);
        for o_1 = 1:n_O
            for o_2 = 1:n_O
                i = find(A(o_1,:) == 1);
                j = find(A(o_1,:) == -1);
                k = find(A(o_2,:) == 1);
                l = find(A(o_2,:) == -1);

                if (~isempty(strfind(kernel, 'gauss')))
	            kernel_mat(o_1, o_2) = (1 + feval(kernel, X(:, i), X(:, k), gauss_kernel_scale) ...
                                              - feval(kernel, X(:, i), X(:, l), gauss_kernel_scale) ...
                                              - feval(kernel, X(:, j), X(:, k), gauss_kernel_scale) ...
                                              + feval(kernel, X(:, j), X(:, l), gauss_kernel_scale)    );
                elseif (~isempty(strfind(kernel, 'poly')))
	            kernel_mat(o_1, o_2) = (1 + feval(kernel, X(:, i), X(:, k), poly_degree, poly_bias) ...
                                              - feval(kernel, X(:, i), X(:, l), poly_degree, poly_bias) ...
                                              - feval(kernel, X(:, j), X(:, k), poly_degree, poly_bias) ...
                                              + feval(kernel, X(:, j), X(:, l), poly_degree, poly_bias)    );
                else
	            kernel_mat(o_1, o_2) = (1 + feval(kernel, X(:, i), X(:, k)) ...
                                              - feval(kernel, X(:, i), X(:, l)) ...
                                              - feval(kernel, X(:, j), X(:, k)) ...
                                              + feval(kernel, X(:, j), X(:, l))    );
                end
       
                if (do_squared_error)
                    if (o_1 == o_2)
                        kernel_mat(o_1, o_2) = kernel_mat(o_1, o_2) + lambda;
                    end
                end
            end
        end
    end
end
toc

% Debugging.
%disp('Kernel matrix:');
%kernel_mat
%disp('Eigen values of kernel matrix:');
%eig(kernel_mat)

% Training
model = kernel_rank_lasvm_sgd_train(X, ...
                                   A, ...
   			                       y, ...
            			           lambda, ...
			                       kernel, ...
            			           n_epochs, ...
             			           tau, ...
            			           max_num_sv, ...
                                   do_squared_error, ...
            			           do_budget, ...
            			           elim_methods{em_idx}, ...
            			           z_normalize, ...
            			           use_sigmoid, ...
            			           compute_obj, ...
            			           plot_bndry, ...
                                   model_mat_file, ...
                                   do_kernel_cache, ...
                                   do_kernel_cache_check, ...
                                   do_point_pairwise_kernel_mat, ...
                                   do_sparse_kernel_mat, ...
                                   do_sv_preallocation, ...
                                   gauss_kernel_scale)

